require 'tempfile'

module TourMigrator
  module Floorplan
    class Report
      attr_reader :svg

      FILE_SIZE_WARN = 300
      FILE_SIZE_ERROR = 1000

      def initialize(svg)
        @svg = svg
        @processed = false
      end

      def size_in_kilobytes
        return @size_in_kilobytes if defined?(@size_in_kilobytes)

        tempfile = Tempfile.new('floorplan-report')
        tempfile.write(svg.to_xml)
        @size_in_kilobytes = tempfile.size.to_f / 1024
      end

      def process
        return if @processed

        process_debug
        process_info
        process_warnings
        process_errors

        @processed = true
      end

      def to_s
        process unless @processed

        report_lines = []

        %i[debug info warnings errors].each do |type|
          if send(type).count.positive?
            report_lines << "*** #{type.to_s.upcase} ***"
            report_lines << send(type).flatten.join("\n")
          end
        end

        report_lines.join("\n")
      end

      def to_h
        process unless @processed

        { debug: debug, info: info, warnings: warnings, errors: errors }
      end

      def hotspot_scene_names
        @hotspot_scene_names ||= hotspots.map { |hotspot| hotspot['id'] }
      end

      private

      def debug
        @debug ||= []
      end

      def info
        @info ||= []
      end

      def warnings
        @warnings ||= []
      end

      def errors
        @errors ||= []
      end

      def hotspots
        @hotspots ||= svg.css('.hotspot')
      end

      def levels
        @levels ||= svg.css('.level')
      end

      def level_switchers
        @level_switchers ||= svg.css('.level-switch')
      end

      def uses
        @uses ||= svg.css('use')
      end

      def linktotours
        @linktotours ||= svg.css('.linktotour')
      end

      def duplicate_links_to_hotspots
        return @duplicate_links_to_hotspots if defined?(@duplicate_links_to_hotspots)

        links_to_hotspots = hotspots.map { |h| h['id'] }.group_by(&:to_s)

        @duplicate_links_to_hotspots = links_to_hotspots.select { |_k, v| v.count > 1 }
      end

      def images
        @images ||= svg.css('image')
      end

      def href_hotspots
        @href_hotspots ||= svg.css('.hotspot').to_a.keep_if { |h| !!h['xlink:href'] }
      end

      def svg_element
        @svg_element ||= svg.css('svg').first
      end

      def linked_tours
        @linked_tours ||= hotspots.map { |hotspot| hotspot['id'].to_s.split('/').first }
      end

      def process_debug
        debug << "Element count: #{elements_count}"
        debug << 'Generated on: ' + Time.now.strftime('%F %H:%M')
      end

      def process_info
        info << 'Hotspots'
        info << "  count: #{hotspots.count}"
        info << '  linked scene names:' if hotspots.count.positive?
        hotspots.each_with_index do |hotspot, index|
          info << '    ' + index.to_s + ': ' + hotspot['id'].to_s
        end

        info << 'LinkToTour'
        info << "  count: #{linktotours.count}"
        info << '  linked tour ids:' if linktotours.count.positive?
        linktotours.each_with_index do |hotspot, index|
          info << '    ' + index.to_s + ': ' + hotspot['id'].to_s
        end

        info << 'Levels'
        info << "  count: #{levels.count}"
        info << '  names:' if levels.count.positive?
        levels.each_with_index do |level, index|
          info << '    ' + index.to_s + ":\t" + level['id'].to_s
        end

        info << 'Level Switchers'
        info << "  count: #{level_switchers.count}"
        info << '  names:' if level_switchers.count.positive?
        level_switchers.each_with_index do |level_switcher, index|
          info << '    ' + index.to_s + ":\t" + level_switcher['href'].to_s
        end
      end

      def process_warnings
        if uses.count.positive?
          warnings << '  <use> usage:'
          uses.map { |hotspot| hotspot['xlink:href'].to_s.strip }
              .uniq
              .each_with_index do |href, index|
            warnings << '    ' + index.to_s + ":\txlink:href=" + href
          end
        end

        # TODO: compares scenes.xml with floorplan's hotspots

        if duplicate_links_to_hotspots.any?
          warnings << '  hotspots: multiple links to same scene id:'
          duplicate_links_to_hotspots.each do |id, hotspot_ids|
            warnings << "    scene id: #{id}, link count: #{hotspot_ids.count}"
          end
        end

        if images.count.positive?
          warnings << '  <image> usage:'
          images.each_with_index do |image, index|
            warnings << '    ' + index.to_s + ":\txlink:href=" + image['xlink:href'].to_s[0..25] + '…'
          end
        end

        warnings << "file size in kB: #{size_in_kilobytes}" if file_size_warning?(size_in_kilobytes)
      end

      def process_errors
        errors << '  <style> element found!' if svg.css('style').count.positive?

        errors << '  not all hotspots nested below a level!' if hotspots_without_level_parent?

        # TODO: check unique names for hotspot for each scene

        # hotspot ath and atv is number (int/float),
        # hotspot has linkedscene or exittocitytour

        if href_hotspots.count.positive?
          errors << '  .hotspot xlink:href:'
          href_hotspots.each_with_index do |hotspot, index|
            errors << '    ' + index.to_s + ":\txlink:href=" + hotspot['xlink:href']
          end
        end

        if svg_element
          if !svg_element['viewBox']
            errors << '  svg missing viewBox attribute!'
            errors << '  svg has width attribute!' if svg_element['width']
            errors << '  svg has height attribute!' if svg_element['height']
          else
            view_box = svg_element['viewBox'].split(' ').map(&:to_i)
            height = view_box[2] - view_box[0]
            width = view_box[3] - view_box[1]

            errors << '  svg dimensions larger than 400px!' if width > 400 || height > 400
            errors << '  svg dimensions are not square!' if width != height
          end

          errors << '  <svg> has id attribute' if svg_element['id']
          errors << '  <svg> has class attribute' if svg_element['class']
        end

        errors << "  file size in kB: #{size_in_kilobytes}" if file_size_error?(size_in_kilobytes)
        errors << '  hotspots linking to different tours!' if linked_tours.uniq.count > 1
      end

      def file_size_warning?(size_in_kilobytes)
        size_in_kilobytes >= FILE_SIZE_WARN && size_in_kilobytes < FILE_SIZE_ERROR
      end

      def file_size_error?(size_in_kilobytes)
        size_in_kilobytes >= FILE_SIZE_ERROR
      end

      def hotspots_without_level_parent?
        return if levels.count.zero?

        svg.css('.level .hotspot').count != hotspots.count
      end

      def elements_count
        svg.css('*').count
      end
    end
  end
end
